@@ -1,5 +1,5 @@
 """
-Handle generation of GRAAL-compatible contact maps from fastq files.
+Handle generation of graal-compatible contact maps from fastq files.
 cmdoret, 20190322
 """
 import os, time, csv, sys, re
@@ -21,6 +21,7 @@ import hicstuff.filter as hcf
 import hicstuff.io as hio
 import hicstuff.distance_law as hcdl
 import matplotlib
+import pathlib
 from hicstuff.version import __version__
 import hicstuff.log as hcl
 from hicstuff.log import logger
@@ -40,13 +41,14 @@ def align_reads(
     """
     Select and call correct alignment method and generate logs accordingly.
     Alignments are filtered so that there at most one alignment per read.
+    Alignments are sorted by read name in the output file.
 
     Parameters
     ----------
     reads : str
         Path to the fastq file with Hi-C reads.
     genome : str
-        Path to the genome bowtie2 index prefix if using bowtie2 or to the 
+        Path to the genome bowtie2/bwa index prefix if using bowtie2 or bwa, or to the 
         fasta if using minimap2.
     out_bam : str
         Path to the output BAM file containing mapped Hi-C reads.
@@ -55,7 +57,7 @@ def align_reads(
     threads : int
         Number of threads to run alignments in parallel.
     aligner : bool
-        Use minimap2 instead of bowtie2.
+        Use minimap2 or bwa instead of bowtie2.
     iterative : bool
         Wether to use the iterative mapping procedure (truncating reads and
         extending iteratively)
@@ -69,8 +71,7 @@ def align_reads(
     """
     if tmp_dir is None:
         tmp_dir = os.getcwd()
-    index = None
-    tmp_sam = out_bam + ".tmp"
+    tmp_bam = out_bam + ".tmp"
 
     if iterative:
         iter_tmp_dir = hio.generate_temp_dir(tmp_dir)
@@ -79,37 +80,40 @@ def align_reads(
             tmp_dir=iter_tmp_dir,
             ref=genome,
             n_cpu=threads,
-            bam_out=tmp_sam,
+            bam_out=tmp_bam,
             min_qual=min_qual,
             aligner=aligner,
             read_len=read_len,
         )
         st.rmtree(iter_tmp_dir)
+        sp.call(
+            "samtools view -F 2048 -h -@ {threads} -O BAM {tmp} -o {out}".format(
+                threads=threads, tmp=tmp_bam, out=out_bam
+            ),
+            shell=True,
+        )
     else:
         if aligner == "minimap2":
             map_cmd = "minimap2 -2 -t {threads} -ax sr {fasta} {fastq} > {sam}"
+        elif aligner == "bwa":
+            map_cmd = "bwa mem -t {threads} -v 1 {index} {fastq} > {sam}"
         else:
-            index = hci.check_bt2_index(genome)
             map_cmd = "bowtie2 --very-sensitive-local -p {threads} -x {index} -U {fastq} > {sam}"
         map_args = {
             "threads": threads,
-            "sam": tmp_sam,
+            "sam": tmp_bam,
             "fastq": reads,
             "fasta": genome,
-            "index": index,
+            "index": genome,
         }
         sp.call(map_cmd.format(**map_args), shell=True)
-
-    # Remove supplementary alignments
-    # TODO: replace sp.call with ps.view command. It currently has a bug
-    # preventing output redirection.
-    sp.call(
-        "samtools view -F 2048 -h -@ {threads} -O BAM -o {out} {tmp}".format(
-            tmp=tmp_sam, threads=threads, out=out_bam
-        ),
-        shell=True,
-    )
-    os.remove(tmp_sam)
+        # Remove supplementary alignments and sort reads by name
+        sp.call("samtools view -F 2048 -h -@ {threads} {tmp} | samtools sort -n -@ {threads} -o {out} -".format(
+                tmp=tmp_bam, threads=threads, out=out_bam
+            ),
+            shell=True,
+        )
+    os.remove(tmp_bam)
 
 
 def bam2pairs(bam1, bam2, out_pairs, info_contigs, min_qual=30):
@@ -131,7 +135,7 @@ def bam2pairs(bam1, bam2, out_pairs, info_contigs, min_qual=30):
     info_contigs : str
         Path to the info contigs file, to get info on chromosome sizes and order.
     min_qual : int
-        Minimum mapping quality required to keep a Hi=C pair.
+        Minimum mapping quality required to keep a Hi-C pair.
     """
     forward = ps.AlignmentFile(bam1, "rb")
     reverse = ps.AlignmentFile(bam2, "rb")
@@ -329,7 +333,9 @@ def pairs2cool(pairs_file, cool_file, bins_file):
     os.remove(bins_tmp)
 
 
-def pairs2matrix(pairs_file, mat_file, fragments_file, mat_fmt="GRAAL", threads=1):
+def pairs2matrix(
+    pairs_file, mat_file, fragments_file, mat_fmt="graal", threads=1, tmp_dir=None
+):
     """Generate the matrix by counting the number of occurences of each
     combination of restriction fragments in a pairs file.
 
@@ -343,17 +349,19 @@ def pairs2matrix(pairs_file, mat_file, fragments_file, mat_fmt="GRAAL", threads=
         Path to the fragments_list.txt file. Used to know total
         matrix size in case some observations are not observed at the end.
     mat_fmt : str
-        The format to use when writing the matrix. Can be GRAAL or bg2 format.
+        The format to use when writing the matrix. Can be graal or bg2 format.
     threads : int
         Number of threads to use in parallel.
+    tmp_dir : str
+        Temporary directory for sorting files. If None given, will use the system default.
     """
     # Number of fragments is N lines in frag list - 1 for the header
     n_frags = sum(1 for line in open(fragments_file, "r")) - 1
     frags = pd.read_csv(fragments_file, delimiter="\t")
 
     def write_mat_entry(frag1, frag2, contacts):
-        """Write a single sparse matrix entry in either GRAAL or bg2 format"""
-        if mat_fmt == "GRAAL":
+        """Write a single sparse matrix entry in either graal or bg2 format"""
+        if mat_fmt == "graal":
             mat.write("\t".join(map(str, [frag1, frag2, n_occ])) + "\n")
         elif mat_fmt == "bg2":
             frag1, frag2 = int(frag1), int(frag2)
@@ -376,7 +384,13 @@ def pairs2matrix(pairs_file, mat_file, fragments_file, mat_fmt="GRAAL", threads=
             )
 
     pre_mat_file = mat_file + ".pre.pairs"
-    hio.sort_pairs(pairs_file, pre_mat_file, keys=["frag1", "frag2"], threads=threads)
+    hio.sort_pairs(
+        pairs_file,
+        pre_mat_file,
+        keys=["frag1", "frag2"],
+        threads=threads,
+        tmp_dir=tmp_dir,
+    )
     header_size = len(hio.get_pairs_header(pre_mat_file))
     with open(pre_mat_file, "r") as pairs, open(mat_file, "w") as mat:
         # Skip header lines
@@ -389,7 +403,7 @@ def pairs2matrix(pairs_file, mat_file, fragments_file, mat_fmt="GRAAL", threads=
         pairs_reader = csv.reader(pairs, delimiter="\t")
         # First line contains nrows, ncols and number of nonzero entries.
         # Number of nonzero entries is unknown for now
-        if mat_fmt == "GRAAL":
+        if mat_fmt == "graal":
             mat.write("\t".join(map(str, [n_frags, n_frags, "-"])) + "\n")
         for pair in pairs_reader:
             # Fragment ids are field 8 and 9
@@ -410,8 +424,8 @@ def pairs2matrix(pairs_file, mat_file, fragments_file, mat_fmt="GRAAL", threads=
         n_nonzero += 1
         n_pairs += 1
 
-    # Edit header line to fill number of nonzero entries inplace in GRAAL header
-    if mat_fmt == "GRAAL":
+    # Edit header line to fill number of nonzero entries inplace in graal header
+    if mat_fmt == "graal":
         with open(mat_file) as mat, open(pre_mat_file, "w") as tmp_mat:
             header = mat.readline()
             header = header.replace("-", str(n_nonzero))
@@ -440,36 +454,37 @@ def full_pipeline(
     genome,
     input1,
     input2=None,
-    enzyme=5000,
+    aligner="bowtie2",
+    centromeres=None,
     circular=False,
-    out_dir=None,
-    tmp_dir=None,
-    plot=False,
+    distance_law=False,
+    enzyme=5000,
+    filter_events=False,
+    force=False,
+    iterative=False,
+    mat_fmt="graal",
     min_qual=30,
     min_size=0,
-    threads=1,
     no_cleanup=False,
-    iterative=False,
-    filter_events=False,
-    prefix=None,
-    start_stage="fastq",
-    mat_fmt="GRAAL",
-    aligner="bowtie2",
+    out_dir=None,
     pcr_duplicates=False,
-    distance_law=False,
-    centromeres=None,
+    plot=False,
+    prefix=None,
     read_len=None,
     remove_centros=None,
+    start_stage="fastq",
+    threads=1,
+    tmp_dir=None,
 ):
     """
     Run the whole hicstuff pipeline. Starting from fastq files and a genome to
     obtain a contact matrix.
-    
+
     Parameters
     ----------
     genome : str
-        Path to the bowtie2 index prefix if using bowtie2 or to the genome in
-        fasta format if using minimap2.
+        Path to the bowtie2/bwa index prefix if using bowtie2/bwa or to the genome 
+        in fasta format if using minimap2.
     input1 : str
         Path to the Hi-C reads in fastq format (forward), the aligned Hi-C reads
         in BAM format, or the pairs file, depending on the value of start_stage.
@@ -503,54 +518,74 @@ def full_pipeline(
         alignment.
     filter_events : bool
         Filter spurious or uninformative 3C events. Requires a restriction enzyme.
+    force : bool
+        If True, overwrite existing files with the same name as output.
     prefix : str or None
-        Choose a common name for output files instead of default GRAAL names.
+        Choose a common name for output files instead of default graal names.
     start_stage : str
         Step at which the pipeline should start. Can be "fastq", "bam", "pairs"
-        or "pairs_idx". With starting from bam allows to skip alignment and start 
+        or "pairs_idx". With starting from bam allows to skip alignment and start
         from named-sorted bam files. With
         "pairs", a single pairs file is given as input, and with "pairs_idx", the
         pairs in the input must already be attributed to fragments and fragment
         attribution is skipped.
     mat_fmt : str
-        Select the output matrix format. Can be either "bg2" for the 
-        bedgraph2 format, "cool" for Mirnylab's cool format, or GRAAL for a
+        Select the output matrix format. Can be either "bg2" for the
+        bedgraph2 format, "cool" for Mirnylab's cool format, or graal for a
         plain text COO format compatible with Koszullab's instagraal software.
     aligner : str
-        Read alignment software to use. Can be either "minimap2" or "bowtie2".
+        Read alignment software to use. Can be either "minimap2", "bwa" or "bowtie2".
     pcr_duplicates : bool
         If True, PCR duplicates will be filtered based on genomic positions.
         Pairs where both reads have exactly the same coordinates are considered
         duplicates and only one of those will be conserved.
     distance_law : bool
-        If True, generates a distance law file with the values of the probabilities 
+        If True, generates a distance law file with the values of the probabilities
         to have a contact between two distances for each chromosomes or arms if the
-        file with the positions has been given. The values are not normalized, or 
+        file with the positions has been given. The values are not normalized, or
         averaged.
     centromeres : None or str
         If not None, path of file with Positions of the centromeres separated by a
-        space and in the same order than the chromosomes. 
+        space and in the same order than the chromosomes.
     read_len : int
         Maximum read length to expect in the fastq file. Optionally used in iterative
         alignment mode. Estimated from the first read by default. Useful if input fastq
         is a composite of different read lengths.
     remove_centros : None or int
         If the distance law is computed, this is the number of kb that will be removed
-        around the centromere position given by in the centromere file. 
+        around the centromere position given by in the centromere file.
     """
     # Check if third parties can be run
-    if aligner in ("bowtie2", "minimap2"):
+    if aligner in ("bowtie2", "minimap2", "bwa"):
         if check_tool(aligner) is None:
             logger.error("%s is not installed or not on PATH", aligner)
+            sys.exit(1)
     else:
-        logger.error("Incompatible aligner software, choose bowtie2 or minimap2")
+        logger.error("Incompatible aligner software, choose bowtie2, minimap2 or bwa")
+        sys.exit(1)
     if check_tool("samtools") is None:
         logger.error("Samtools is not installed or not on PATH")
+        sys.exit(1)
 
     # Pipeline can start from 3 input types
     start_time = datetime.now()
     stages = {"fastq": 0, "bam": 1, "pairs": 2, "pairs_idx": 3}
     start_stage = stages[start_stage]
+
+    # Check if the number of input files is correct
+    if start_stage <= 1:
+        if input2 is None:
+            logger.error(
+                "You must provide 2 input files when --start-stage is fastq " "or bam."
+            )
+            sys.exit(1)
+    else:
+        if input2 is not None:
+            logger.error(
+                "You must provide a single input file when --start-stage is "
+                "pairs or pairs_idx."
+            )
+            sys.exit(1)
     # sanitize enzyme
     enzyme = str(enzyme)
     # Remember whether fragments_file has been generated during this run
@@ -584,12 +619,27 @@ def full_pipeline(
     def _tmp_file(fname):
         if prefix:
             fname = prefix + "." + fname
-        return join(tmp_dir, fname)
+        full_path = join(tmp_dir, fname)
+        if not force and os.path.exists(full_path):
+            raise IOError(
+                "Temporary file {} already exists. Use --force to overwrite".format(
+                    full_path
+                )
+            )
+        return full_path
 
     def _out_file(fname):
         if prefix:
             fname = prefix + "." + fname
-        return join(out_dir, fname)
+        full_path = join(out_dir, fname)
+        if not force and os.path.exists(full_path):
+            raise IOError(
+                "Output file {} already exists. Use --force to overwrite".format(
+                    full_path
+                )
+            )
+
+        return full_path
 
     # Define temporary file names
     log_file = _out_file("hicstuff_" + now + ".log")
@@ -601,44 +651,95 @@ def full_pipeline(
     pairs_filtered = _tmp_file("valid_idx_filtered.pairs")
     pairs_pcr = _tmp_file("valid_idx_pcrfree.pairs")
 
+    # Enable file logging
+    hcl.set_file_handler(log_file)
+    generate_log_header(log_file, input1, input2, genome, enzyme)
+
     # If the user chose bowtie2 and supplied an index, extract fasta from it
     # For later steps of the pipeline (digestion / frag attribution)
-    if aligner == "bowtie2":
+    # Check if the genome is an index or fasta file
+    idx = hio.check_fasta_index(genome, mode=aligner)
+    is_fasta = hio.check_is_fasta(genome)
+    
+    # Different aligners accept different files. Make sure the input format is good.
+    # Note bowtie2 can extract fasta from the index, but bwa cannot
+    sane_input = {
+            'bowtie2': is_fasta or idx,
+            'minimap2': is_fasta, 
+            'bwa': is_fasta
+    }
+
+    if not sane_input[aligner]:
+        logger.error("You must provide either a fasta or bowtie2 index prefix as genome")
+
+    # Just use the input genome is it is indexed
+    if is_fasta and idx:
+        fasta = genome
+    # Otherwise copy it in tmpdir for indexing, unless the input is a bt2 index, in which
+    # case fasta will be extracted later from it.
+    else:
+        if is_fasta:
+            st.copy(genome, tmp_genome)
+            genome = tmp_genome
+        fasta = tmp_genome
+        
+
+    # Bowtie2-specific feature: extract fasta from the index
+    if aligner == 'bowtie2' and not is_fasta:
+        # Index is present, extract fasta file from it
         bt2fa = sp.Popen(
-            ["bowtie2-inspect", genome], stdout=open(tmp_genome, "w"), stderr=sp.PIPE
+            ["bowtie2-inspect", genome],
+            stdout=open(tmp_genome, "w"),
+            stderr=sp.PIPE,
         )
         _, bt2err = bt2fa.communicate()
         # bowtie2-inspect still has return code 0 when crashing, need to
         # actively look for error in stderr
-        fasta = tmp_genome
         if re.search(r"[Ee]rror", bt2err.decode()):
+
             logger.error(bt2err)
             logger.error(
                 "bowtie2-inspect has failed, make sure you provided "
                 "the path to the bowtie2 index without the extension."
             )
             sys.exit(1)
-    else:
-        fasta = genome
 
-    # Enable file logging
-    hcl.set_file_handler(log_file)
-    generate_log_header(log_file, input1, input2, genome, enzyme)
+    # Build index with bowtie2 / bwa if required
+    if idx is None and aligner in ['bowtie2', 'bwa']:
+        if aligner == 'bowtie2':
+            index_cmd = ["bowtie2-build", '-q', fasta, fasta]
+        elif aligner == 'bwa':
+            index_cmd = ['bwa', 'index', fasta]
+        # We only need the index if the user provided fastq input
+        if start_stage == 0:
+            # If no index present assume input is fasta, copy it in tmp and
+            # index it (to avoid conflict between instances)
+            logger.info(
+                "%s index not found at %s, generating "
+                "a local temporary index.", aligner, genome
+            )
+            sp.run(index_cmd, stderr=sp.PIPE)
 
+    # Check for spaces in fasta headers and issue error if found
+    for record in SeqIO.parse(fasta, "fasta"):
+        if " " in record.id:
+            logger.error(
+                "Sequence identifiers contain spaces. Please clean the input genome."
+            )
     # Define output file names (tsv files)
     if prefix:
         fragments_list = _out_file("frags.tsv")
         info_contigs = _out_file("chr.tsv")
         mat = _out_file("mat.tsv")
         # If matrix has a different format, give it the right extension
-        if mat_fmt != "GRAAL":
+        if mat_fmt != "graal":
             mat = _out_file(mat_fmt)
     else:
-        # Default GRAAL file names
+        # Default graal file names
         fragments_list = _out_file("fragments_list.txt")
         info_contigs = _out_file("info_contigs.txt")
         mat = _out_file("abs_fragments_contacts_weighted.txt")
-        if mat_fmt != "GRAAL":
+        if mat_fmt != "graal":
             mat = _out_file("abs_fragments_contacts_weighted." + mat_fmt)
     # Define what input files are given
     if start_stage == 0:
@@ -677,11 +778,6 @@ def full_pipeline(
             min_qual=min_qual,
             read_len=read_len,
         )
-        # Sort alignments by read name
-        ps.sort("-@", str(threads), "-n", "-O", "BAM", "-o", bam1 + ".sorted", bam1)
-        st.move(bam1 + ".sorted", bam1)
-        ps.sort("-@", str(threads), "-n", "-O", "BAM", "-o", bam2 + ".sorted", bam2)
-        st.move(bam2 + ".sorted", bam2)
 
     # Starting from bam files
     if start_stage <= 1:
@@ -722,6 +818,7 @@ def full_pipeline(
         pairs_idx + ".sorted",
         keys=["chr1", "pos1", "chr2", "pos2"],
         threads=threads,
+        tmp_dir=tmp_dir,
     )
     os.rename(pairs_idx + ".sorted", pairs_idx)
 
@@ -759,7 +856,7 @@ def full_pipeline(
         if remove_centros is None:
             remove_centros = 0
         remove_centros = int(remove_centros)
-        x_s, p_s, names_distance_law = hcdl.get_distance_law(
+        x_s, p_s, _ = hcdl.get_distance_law(
             pairs_idx,
             fragments_list,
             centro_file=centromeres,
@@ -789,7 +886,14 @@ def full_pipeline(
         cool_file = os.path.splitext(mat)[0] + ".cool"
         pairs2cool(use_pairs, cool_file, fragments_list)
     else:
-        pairs2matrix(use_pairs, mat, fragments_list, mat_fmt=mat_fmt, threads=threads)
+        pairs2matrix(
+            use_pairs,
+            mat,
+            fragments_list,
+            mat_fmt=mat_fmt,
+            threads=threads,
+            tmp_dir=tmp_dir,
+        )
 
     # Clean temporary files
     if not no_cleanup:
